; -*- indent-tabs-mode: t; -*-
BITS 16

ORG 0x7c00

%define doublebuffer
; %define vsync
; %define debuglogo

struc datapool
  scratch  resd  1
  maglimit resw  1
  xoff     resd  1
  yoff     resd  1
  scale    resd  1
  sv       resd  1
  iox      resd  1
  ioy      resd  1
  iow      resd  1
  iowp     resw  1
  iohp     resw  1
  l100     resw  1
  l160     resw  1
  l16      resw  1
  iterations     resb  1
endstruc
%define ioh iow

%define cLimit        0.006
%define cIterations   100

	mov ax, 0x13
	int 0x10
%ifdef doublebuffer
	push 0xA000
	pop es
	push 0x8000
	pop fs
%else
	push 0xA000
	pop fs
%endif

	mov dx, 0x03c8
	xor ax, ax
	xor cx, cx
	out dx, al
	inc dx
setpal:
	mov al, 3
	mul cl
	out dx, al
	mov al, cl
	out dx, al
	mov al, 7
	mul cl
	out dx, al
	loop setpal

	mov di, data

	cld
	fninit

	fnstcw word [di]
	or     word [di], 1<<10
	fldcw  word [di]

draw_one:
	mov bx, 0xfa00

loop:
	fild word [di+maglimit]
	;; Compute the x and y pixel coordinates from the counter in bx
	;; x = bx % 320
	;; y = bx / 320
	;; Fortunately, 'div' computes both at once for us
	mov ax, bx
	dec ax
	xor dx, dx
	mov cx, 320
	div cx
	;; ax = y, dx = x

	;; Compute the floating-point x and y coordinates from the
	;; pixel coordinates
	sub ax, 100
	mov word [di], ax
	fild word [di]
	fidiv word [di+l100]
	fmul dword [di+scale]
	fadd dword [di+yoff]

	sub dx, 160
	mov word [di], dx
	fild word [di]
	fidiv word [di+l160]
	fmul dword [di+scale]
	fadd dword [di+xoff]

	;; ST(0) = x
	;; ST(1) = y

	;; Check if we're inside the IO logo
	;; check Y
	fld  st1
	fsub dword [di+ioy]
	fabs
	fld  dword [di+ioh]
	fcomip st1
	fstp st4
	jc outio

	;; check X
	fld  st0
	fsub dword [di+iox]
	fabs
	fld  dword [di+iow]
	fcomip st1
	fstp st4
	jnc inio

outio:
	;; We're outside the IO logo.
	;; Do the mandelbrot set thing.
	;; See <http://en.wikipedia.org/wiki/Mandelbrot_set> for more details

	;; We use the top four elements of the x87 floating-point ;;
	;; stack across iterations of the inner loop. These are used to
	;; store z_n and c, which are both complex numbers.

	;; z_{n+1} = z_n^2 + c
	;; z   = (R,I)   = (ST(0), ST(1))
	;; c   = (cr,ci) = (ST(2), ST(3))
	;; z^2 = (R^2 - I^2, 2RI)
	;; z'  = (R^2 - I^2 + cr, 2RI + ci)


	;; Initialize z_0 = (0,0)
	fldz
	fldz

	mov dx, cIterations
iter:
	fld   st0
	fmul  st2
	fadd  st0
	fadd  st4
	;; ST(0) = 2RI + Ci
	fld   st2
	fmulp st3
	;; stack = (C', R, I^2, Cr, Ci)
	fld   st1
	fmul  st2
	fsub  st3
	fadd  st4
	;; stack = (R', C', R, I^2, Cr, Ci)
	fstp  st2
	fstp  st2

	;; Compute magnitude^2 = r^2+c^2
	fld   st0
	fmul  st0
	fld   st2
	fmul  st0
	faddp st1

	;; If magnitude^2 > a specified limit, we've escaped, break out of the loop.
	;; (mag, R, C, Cr, Ci, lim)
	fcomip st5
	jnc escape

	dec dx
	jnz iter
escape:

	test dl,dl
	jz dopop

	;; If this point wasn't in the set, add 32 to the number of
	;; iterations remaining to get the color.
	;;
	;; 32 is an arbitrary choice based on the default
	;; palette. Play with other formulas here to get a different
	;; color scheme.

	add dl, 32
dopop:
	;; Discard the dead items from the stack
	mov cl,5
fpop:
	fstp st1
	loop fpop

drawpx:
	;; Draw the pixel
	mov byte [fs:bx - 1], dl

	dec bx
	jnz loop

%ifdef doublebuffer
%ifdef vsync
	;; Refresh the screen. First, wait for the vblank interval.
	mov dx, 0x3da
vblank:
	in al, dx
	test al, 0x8
	jnz vblank
waitvb:
	in al, dx
	test al, 0x8
	jz waitvb
%endif

	;; Now swap the double buffer, using 'rep movsw' to do the
	;; copy. movsw copies from ds:si to es:di, so we need to
	;; temporarily swap ds to point at our double-buffer region.
	push ds
	push fs
	pop  ds
	push di
	xor  si, si
	xor  di, di
	mov  cx, 320*200/2
	rep  movsw
	pop  di
	pop  ds
%endif

	;; Zoom in. multiply "scale" by "sv" to get the new scale
	;; factor. If the resulting value is smaller than 'limit', we're done.

	fld    dword [di+scale]
	fmul   dword [di+sv]
	fst    dword [di+scale]
	dec    byte  [di+iterations]
	jz hcf

	jmp draw_one

hcf: jmp hcf

inio:
	push bx

	;; compute X
	fsub  dword [di+iox]
	fimul word  [di+iowp]
	fdiv  dword [di+iow]
	fist  word  [di]
	fisub word  [di]
	mov   cx, word [di]

	;; Y
	fxch  st1
	fsub  dword [di+ioy]
	fimul word [di+iohp]
	fdiv  dword [di+ioh]
	fist  word  [di]
	fisub word  [di]
	mov   si, word [di]

	;; Stack: (y, x, lim)

	;; Now index into the bitmap to find the byte at the current
	;; position.

	shl si, 1
	mov ax, word [si + iodata + 12]
	mov bx, word [si + iodata + 14]

	;; (y, x, lim)
	fld1
	fsub st0, st2
	;; (1-x, y, x, lim)
	fld1
	fsub st0, st2

	fldz
	;; Stack: ([scratch], acc, 1-y, 1-x, y, x, lim)

	bt ax, cx
	jnc noA

	fadd  st1
	fmul  st2
noA:

	bt bx, cx
	jnc noC
	fld   st2
	fmul  st4
	faddp st1
noC:

	inc cl

	bt ax, cx
	jnc noB
	fld   st1
	fmul  st5
	faddp st1
noB:

	bt bx, cx
	jnc noD
	fld   st3
	fmul  st5
	faddp st1
noD:

	pop bx
	fimul word [di+l16]
	fistp word [di]
	mov   dx, word [di]
	test dl, dl
	jz dopop
	add dl, 61

	jmp dopop

iodata:
%include "sts.img"
	dw 0

data:
	dd 0			; Scratch space

	;; These are a bunch of constants that determine our
	;; operation. You can change these to change the animation.

	;; How large we let z grow before aborting. Mostly boring, but
	;; feel free to play with it.
	dw 25			; maglimit

	;; The (x,y) coordinates the animation centers on
	dd -0.656		; xoff
	dd -0.450			; yoff

	;; The initial zoom factor
%ifdef debuglogo
	dd cLimit		; scale
%else
	dd 2.0			; scale
%endif
	;; How much to zoom in every frame
	dd 0.795			; sv
	;; When to stop zooming
	;; The coordinates to draw the IO logo at
	dd -0.6587		; iox
	dd -0.450		; ioy
	;; The width and height of the IO logo
	dd 0.000625		; iodim

	dw 8			; iowp
	dw 6			; iohp
	dw 100
	dw 160
	dw 16
%ifdef debuglogo
	db 1
%else
	db 28
%endif

;padding: times 446 - ($-$$) db 0x90
;partitiontable: times 64 db 0xff
;signature: db 0x55, 0xaa
